﻿# =============================================================================
# >> IMPORTS
# =============================================================================
# Python
from __future__ import with_statement

import binascii
import os
import re
import time

from collections import defaultdict
from collections import deque
from configobj   import ConfigObj
from path        import path

# EventScripts
import es
import cfglib

# Source-Python-Extensions
import spe

from spe import HookType
from spe import HookAction
from spe import Conventions


# =============================================================================
# >> ADDON INFORMATION
# =============================================================================
info = es.AddonInfo()
info.name        = 'Tidy Chat'
info.version     = '1.1'
info.author      = 'Ayuto'
info.url         = 'http://addons.eventscripts.com/addons/user/53757'
info.basename    = 'tidychat'
info.description = 'An addon to keep the chat tidy.'

es.ServerVar(info.basename + '_version', info.version).makepublic()


# =============================================================================
# >> GLOBAL VARIABLES
# =============================================================================
ADDON_PATH = path(__file__).parent
DATA_PATH  = ADDON_PATH.joinpath('data')
DATA       = ConfigObj(DATA_PATH.joinpath('data.ini'))
players    = defaultdict(deque)


# =============================================================================
# >> CONFIGURATION
# =============================================================================
cfg = cfglib.AddonCFG(ADDON_PATH.joinpath('settings.cfg'))

cfg.text('*************')
cfg.text('  Tidy Chat')
cfg.text('*************')

blocked_events = cfg.cvar(
    'tc_blocked_events',
    'player_connect; player_disconnect; player_team; server_cvar',
    'Events, which are not sent to the clients to prevent messages etc.')

message_count = cfg.cvar(
    'tc_message_count',
    '2',
    'Number of messages a player is allowed to sent in "tc_message_time" ' + \
        'seconds')

message_time = cfg.cvar(
    'tc_message_time',
    '3',
    'Seconds until the number of messages are reseted')

replacement = cfg.cvar(
    'tc_replacement',
    '!@#$&',
    'Replacement for bad words')

radio_chat = cfg.cvar(
    'tc_radio_chat',
    '1',
    '0 = As usal\n// 1 = No radio chat message\n// 2 = No radio sound/mes' + \
        'sage')

client_print = cfg.cvar(
    'tc_client_print',
    '#;',
    'Block messages, which start with one of these phrases')

debug = cfg.cvar(
    'tc_debug',
    '0',
    'Enables/disables debug messages (useful for configurating this addon)')

cfg.write()
cfg.execute()


# =============================================================================
# >> CLASSES
# =============================================================================
class VFunc(object):
    def __init__(self, name, ptr):
        data = DATA['vtable'][name]
        self.ptr    = ptr
        self.offset = int(data['offset'])
        self.params = data['params']
        self.func   = spe.findVirtualFunc(ptr, self.offset)

    def __call__(self, *args):
        spe.setCallingConvention('thiscall')
        return spe.callFunction(self.func, self.params, (self.ptr,) + args)

    def hook(self, callback, htype=HookType.Pre):
        spe.hookFunction(self.func, self.params, Conventions.THISCALL, htype,
            callback)

    def unhook(self, callback, htype=HookType.Pre):
        spe.unHookFunction(self.func, htype, callback)


# =============================================================================
# >> LOAD & UNLOAD
# =============================================================================
def load():
    spe.parseINI(ADDON_PATH.namebase + '/data' + '/signatures.ini')
    spe.detourFunction('CSRadioMessage', HookType.Pre, preRadioMessage)
    spe.detourFunction('ClientPrintFilter', HookType.Pre, preClientPrint)
    FIRE_EVENT_FUNC.hook(preFireEvent)
    es.addons.registerSayFilter(sayFilter)

def unload():
    spe.undetourFunction('CSRadioMessage', HookType.Pre, preRadioMessage)
    spe.undetourFunction('ClientPrintFilter', HookType.Pre, preClientPrint)
    FIRE_EVENT_FUNC.unhook(preFireEvent)
    es.addons.unregisterSayFilter(sayFilter)


# =============================================================================
# >> GAME EVENTS
# =============================================================================
def player_disconnect(ev):
    players.pop(int(ev['userid']), 0)


# =============================================================================
# >> FUNCTIONS
# =============================================================================
def dbgmsg(msg):
    if int(debug):
        es.dbgmsg(0, '[Tidy Chat] ' + msg)

def getPointer(ptrname):
    if os.name == 'nt':
        data   = DATA['pointers'][ptrname]
        sig    = binascii.unhexlify(data['sig'].replace(' ', ''))
        offset = int(data['offset'])
        return spe.getLocVal('i', spe.getLocVal('i',
            spe.getPointer(sig, offset)))

    return spe.getLocVal('i', spe.findSymbol(ptrname))

def getBlacklist():
    def filterLine(line):
        return not (line.startswith('//') or not line)

    with ADDON_PATH.joinpath('blacklist.txt').open() as f:
        for pattern in filter(filterLine, f.read().split('\n')):
            yield re.compile(re.escape(pattern), re.IGNORECASE)


# =============================================================================
# >> CALLBACKS
# =============================================================================
def preClientPrint(args):
    dbgmsg('Blockable message: ' + args[2])
    for msg in map(lambda x: x.strip(' '), str(client_print).split(';')):
        if not msg or args[2].startswith(msg):
            continue

        for userid in es.getUseridList():
            spe.call('RemoveRecipient', args[0], spe.getPlayer(userid))

        break

    return (HookAction.Continue, 0)

def preRadioMessage(args):
    value = int(radio_chat)
    if value == 1:
        # Make sure that all string arguments are really strings. Otherwise
        # an error would raise
        for x in xrange(3, 8):
            args[x] = ''

        return (HookAction.Modified, 0)

    if value == 2:
        for userid in es.getUseridList():
            spe.call('RemoveRecipient', args[0], spe.getPlayer(userid))

    return (HookAction.Continue, 0)

def preFireEvent(args):
    event = VFunc('IGameEvent::GetName', args[0])()
    dbgmsg('Blockable event: ' + event)
    if event not in str(blocked_events):
        return (HookAction.Continue, 0)

    args[1] = True
    return (HookAction.Modified, 0)

def sayFilter(userid, text, teamonly):
    # Check spam
    now    = time.time()
    player = players[userid]
    while player and player[0] < now:
        player.popleft()

    if len(player) > int(message_count):
        es.tell(userid,
            'You need to wait %.1f seconds until you can chat again.'% (
                player[0] - now))

        return (0, 0, 0)

    player.append(now + int(message_time))

    # Check blacklist
    text = text.strip('"')
    for pattern in BLACKLIST:
        text = pattern.sub(str(replacement), text)

    return (userid, text, teamonly)


# =============================================================================
# >> GLOBAL VARIABLES
# =============================================================================
gameeventmanager = getPointer('gameeventmanager')
FIRE_EVENT_FUNC  = VFunc('IGameEventManager2::FireEvent', gameeventmanager)
BLACKLIST        = tuple(getBlacklist())